#include "fff_iconic_match.h"
#include "fff_base.h"
#include "fff_cubic_spline.h" 

#include <math.h>
#include <stdlib.h>
#include <string.h>


#define SPEEDUP_REF 16777216   /* 256^3 */ 

#define APPLY_AFFINE_TRANSFORMATION(Tx,Ty,Tz,Tvox,bufTvox,x,y,z)   \
  bufTvox = Tvox;						   \
  Tx = (*bufTvox)*x; bufTvox++;					   \
  Tx += (*bufTvox)*y; bufTvox++;				   \
  Tx += (*bufTvox)*z; bufTvox++;				   \
  Tx += *bufTvox; bufTvox++;					   \
  Ty = (*bufTvox)*x; bufTvox++;					   \
  Ty += (*bufTvox)*y; bufTvox++;				   \
  Ty += (*bufTvox)*z; bufTvox++;				   \
  Ty += *bufTvox; bufTvox++;					   \
  Tz = (*bufTvox)*x; bufTvox++;					   \
  Tz += (*bufTvox)*y; bufTvox++;				   \
  Tz += (*bufTvox)*z; bufTvox++;				   \
  Tz += *bufTvox


static double _fff_imatch_marginalize( double* h, const double* H, int clampI, int clampJ, int axis ); 
static void _fff_L1_moments ( const double * h, int clamp, int stride, double* median, double* dev, double* sumh );


/* =========================================================================
   JOINT HISTOGRAM COMPUTATION 
   ========================================================================= */

/*
  
imI : assumed signed short encoded, possibly non-contiguous.

imJ_padded : assumed C-contiguous (last index varies faster) & signed
short encoded.

H : assumed C-contiguous 

Tvox : assumed C-contiguous

Negative intensities are ignored. 

*/

void fff_imatch_joint_hist( double* H, int clampI, int clampJ,  
			    const fff_array* imI,
			    const fff_array* imJ_padded, 
			    const double* Tvox )
{
  const signed short* bufI=(signed short*)imI->data; 
  fff_array_iterator iterI = fff_array_iterator_init(imI); 
  size_t p = 0; 
  const signed short* bufJ=(signed short*)imJ_padded->data; 
  size_t dimJX=imJ_padded->dimX-2, dimJY=imJ_padded->dimY-2, dimJZ=imJ_padded->dimZ-2;  
  size_t Q[8]; 
  double W[8]; 
  size_t *bufQ;
  double *bufW; 
  int i, j, k, clampJ_i;
  size_t q;
  size_t u2=imJ_padded->dimZ; 
  size_t u3=u2+1; 
  size_t u4=(imJ_padded->dimY)*u2;
  size_t u5=u4+1; 
  size_t u6=u4+u2; 
  size_t u7=u6+1; 
  size_t x, y, z; 
  int nx, ny, nz;
  double Tx, Ty, Tz; 
  const double* bufTvox; 
  double wx, wy, wz, wxwy, wxwz, wywz, wxwywz; 

  /* Re-initialize joint histogram */ 
  memset( (void*)H, 0, clampI*clampJ*sizeof(double) );

  /* Looop over source voxels */
  while( iterI.uncomplete ) {
  
    /* Source voxel intensity */
    i = bufI[p]; 

    /* Source voxel coordinates */
    x = iterI.x;
    y = iterI.y;
    z = iterI.z;
    
    /* Compute the transformed grid coordinates of current voxel */ 
    APPLY_AFFINE_TRANSFORMATION(Tx, Ty, Tz, Tvox, bufTvox, x, y, z);     
    
    /* Test whether the current voxel is below the intensity
       threshold, or the transformed point is completly outside
       the reference grid */
    if ( (i>=0) &&
	 (Tx>-1) && (Tx<dimJX) && 
	 (Ty>-1) && (Ty<dimJY) && 
	 (Tz>-1) && (Tz<dimJZ) ) {
      
      /* 
	 Nearest neighbor (floor coordinates in the padded
	 image, hence +1). 
	 
	 Notice that using the floor function doubles excetution time.
	 
	 FIXME: see if we can replace this with assembler instructions. 
      */
      nx = FFF_FLOOR(Tx) + 1;
      ny = FFF_FLOOR(Ty) + 1;
      nz = FFF_FLOOR(Tz) + 1;
      
      /* The convention for neighbor indexing is as follows:
       *
       *   Floor slice        Ceil slice
       *
       *     2----6             3----7                     y          
       *     |    |             |    |                     ^ 
       *     |    |             |    |                     |
       *     0----4             1----5                     ---> x
       */
      
      /* Compute index matches and prior weights */ 
      
      /*** Trilinear interpolation weights.  
	   Note: wx = nnx + 1 - Tx, where nnx is the location in
	   the NON-PADDED grid */ 
      wx = nx - Tx; 
      wy = ny - Ty;
      wz = nz - Tz;
      wxwy = wx*wy;    
      wxwz = wx*wz;
      wywz = wy*wz;
      wxwywz = wxwy*wz; 
      
      /*** Prepare buffers */ 
      bufQ = Q; 
      bufW = W; 
      
      /*** Neighbor 0: (0,0,0) */ 
      q = nx*u4 + ny*u2 + nz; 
      *bufQ = q; bufQ++;  
      *bufW = wxwywz; bufW ++; 
      
      /*** Neighbor 1: (0,0,1) */ 
      *bufQ = q + 1; bufQ++;  
      *bufW = wxwy-wxwywz; bufW ++;  
      
      /*** Neighbor 2: (0,1,0) */ 
      *bufQ = q + u2; bufQ++;  
      *bufW = wxwz-wxwywz; bufW ++;  
      
      /*** Neightbor 3: (0,1,1) */ 
      *bufQ = q + u3; bufQ++;  
      /* *bufW = wx-wxwy-wxwz+wxwywz; bufW ++;   */
      *bufW = wx-wxwy-W[2]; bufW ++; 
      
      /*** Neighbor 4: (1,0,0) */ 
      *bufQ = q + u4; bufQ++;  
      *bufW = wywz-wxwywz; bufW ++; 
      
      /*** Neighbor 5: (1,0,1) */ 
      *bufQ = q + u5; bufQ++;  
      /* *bufW = wy-wxwy-wywz+wxwywz; bufW ++;  */
      *bufW = wy-wxwy-W[4]; bufW ++;  
      
      /*** Neighbor 6: (1,1,0) */ 
      *bufQ = q + u6; bufQ++;  
      /* *bufW = wz-wxwz-wywz+wxwywz; bufW ++;  */
      *bufW = wz-wxwz-W[4]; bufW ++; 
      
      /*** Neighbor 7: (1,1,1) */ 
      *bufQ = q + u7; bufQ++;  
      /* *bufW = 1-wx-wy-wz+wxwy+wxwz+wywz-wxwywz; bufW ++;  */
      *bufW = 1-W[3]-wy-wz+wywz; bufW ++; 
      
      /* Here we could update W as the posterior probabilities
	 of the match given the target, the transformation and
	 the intensity distribution */ 
      /*** Do nothing for now */ 
      
      /* For each match, increment the joint histogram by match weight
	 at corresponding entry */
      bufQ = Q; 
      bufW = W; 
      clampJ_i = clampJ*i;
      for(  k=0; k<8; k++, bufQ++, bufW++ ) {
	q = *bufQ; 
	j = bufJ[q]; 
	if (j>=0) 
	  H[j+clampJ_i] += *bufW;
      }
      
    } /* End of IF TRANSFORMS INSIDE */
    
    /* Update source index */ 
    p += fff_array_iterator_update(&iterI); 
    
  } /* End of loop over voxels */ 
  
  return; 
}


/*
  axis == 0 -> source histogram
  axis == 1 -> target histogram 
  
  returns the sum of H 

*/
static double _fff_imatch_marginalize( double* h, const double* H, int clampI, int clampJ, int axis )
{
  int i, j; 
  const double *bufH = H; 
  double *bufh = h;
  double hij, sumH = 0.0; 
  
  if ( axis == 0 ) {
    memset( (void*)h, 0, clampI*sizeof(double) );
    for (i=0; i<clampI; i++, bufh++) 
      for (j=0; j<clampJ; j++, bufH++) {
	hij = *bufH; 
	sumH += hij; 
	*bufh += hij; 
      }
  }
  else if ( axis == 1 ) {
    memset( (void*)h, 0, clampJ*sizeof(double) );
    for (j=0; j<clampJ; j++, bufh++) {
      bufH = H + j; 
      for (i=0; i<clampI; i++, bufH+=clampJ) {
	hij = *bufH; 
	sumH += hij; 
	*bufh += hij; 
      }
    }
  }
  
  return sumH; 
}




/* =========================================================================
   HISTOGRAM-BASED SIMILARITY MEASURES 
   ========================================================================= */

#define TINY 1e-5
#define NICELOG(x)((x)>(TINY) ? log(x):log(TINY)) 

static double _fff_imatch_cc( const double* H, int clampI, int clampJ, double* n ); 
static double _fff_imatch_cr( const double* H, int clampI, int clampJ, double* n ); 
static double _fff_imatch_crL1( const double* H, double* hI, int clampI, int clampJ, double* n ); 
static double _fff_imatch_entropy( const double* h, size_t size, double* n ); 
static double _fff_imatch_mi( const double* H, double* hI, int clampI, double* hJ, int clampJ, double* n ); 
static double _fff_imatch_woe( const double* H, const double* F, 
			       double* fI, int clampI, double* fJ, int clampJ, 
			       double* n ); 

double fff_imatch_cc( const double* H, int clampI, int clampJ )
{
  double n; 
  return( _fff_imatch_cc(H, clampI, clampJ, &n) );
}

double fff_imatch_max_woe_cc( const double* H, int clampI, int clampJ )
{
  double n, x;  
  x = 1 - _fff_imatch_cc(H, clampI, clampJ, &n);
  return( -.5 * n * NICELOG(x) );
}


static double _fff_imatch_cc( const double* H, int clampI, int clampJ, double* n )
{
  int i, j;
  double CC, mj, mi, mij, mj2, mi2, varj, vari, covij, na;
  double aux, auxj, auxi;
  const double *buf;

  /* Initialization */
  na = 0;
  mj  = 0;
  mi  = 0;
  mij = 0;
  mj2 = 0;
  mi2 = 0;
  
  /* Computes conditional moments */
  buf = H;
  for (i=0; i<clampI; i++) 
    for (j=0; j<clampJ; j++, buf++) {
      
      aux = *buf; /* aux == H(i,j) */
      auxj = j * aux;
      auxi = i * aux;
      
      na += aux;
      mj  += auxj;
      mi  += auxi;
      mj2 += j * auxj;
      mi2 += i * auxi;
      mij += i * auxj;
      
  }
  
  /* Test if the dataset is empty */
  if ( na <= 0 ) {
    *n = 0; 
    return 0;
  }
  
  /* Normalization */
  mj  = mj / na;
  mi  = mi / na;
  mj2 = mj2 / na;
  mi2 = mi2 / na;
  mij = mij / na;
  
  /* Compute covariance and variance */
  covij = mij - mj * mi;
  covij = FFF_SQR(covij);
  varj  = mj2 - mj * mj;
  vari  = mi2 - mi * mi;
  aux = varj * vari;
  
  /* Check the variance product is non zero */
  if ( aux <= 0 ) {
    *n = na;
    return 0;
  }
  
  /* Return the correlation coefficient */
  CC = covij / aux;
  
  *n = na; 
  return CC;
     
}


double fff_imatch_cr( const double* H, int clampI, int clampJ )
{
  double n; 
  return( _fff_imatch_cr(H, clampI, clampJ, &n) );
}

double fff_imatch_max_woe_cr( const double* H, int clampI, int clampJ )
{
  double n, x;  
  x = 1 - _fff_imatch_cr(H, clampI, clampJ, &n);
  return( -.5 * n * NICELOG(x) );
}


static double _fff_imatch_cr( const double* H, int clampI, int clampJ, double* n )
{
  int i, j;
  double CR, na, mean, var, cvar, nJ, mJ, vJ, aux, aux2;
  const double *buf;

  /* Initialization */
  na = mean = var = cvar = 0;

  /* Compute conditional moments */
  for (j=0; j<clampJ; j++) {
    nJ = mJ = vJ = 0;
    buf = H + j;
    for (i=0; i<clampI; i++, buf+=clampJ) {
      aux = *buf; /* aux == H(j,i) */
      nJ += aux;
      aux2 = i*aux;
      mJ += aux2;
      vJ += i*aux2;
    }
    if (nJ>0) {
      na += nJ;
      mean += mJ;
      var += vJ;
      mJ /= nJ;
      vJ = vJ/nJ - mJ*mJ;
      cvar += nJ*vJ;
    }
  }
    
  /* Compute total moments */
  if ( na <= 0 ) {
    *n = 0;
    return 0;
  }

  
  mean /= na;
  var = var/na - mean*mean;
  cvar /= na;

  /* Test on total variance */
  if (var<=0) {
    *n = na; 
    return 0;
  }
  
  /* Correlation ratio */
  CR = 1 - cvar/var;

  return CR;
  
}


double fff_imatch_crL1( const double* H, double* hI, int clampI, int clampJ )
{
  double n; 
  return( _fff_imatch_crL1(H, hI, clampI, clampJ, &n) );
}

double fff_imatch_max_woe_crL1( const double* H, double* hI, int clampI, int clampJ )
{
  double n, x;  
  x = 1 - _fff_imatch_crL1(H, hI, clampI, clampJ, &n);
  return( -.5 * n * NICELOG(x) );
}


static double _fff_imatch_crL1( const double* H, double* hI, int clampI, int clampJ, double* n )
{
  int j;
  double na, med, dev, cdev, nJ, mJ, dJ;
   
  /* Initialization */
  na = cdev = 0;

  /* Compute conditional moments */  
  for (j=0; j<clampJ; j++) {
    _fff_L1_moments ( H+j, clampI, clampJ, &mJ, &dJ, &nJ );
    cdev += nJ*dJ;
    na += nJ;
  }
  
  /* Conditional dispersion */
  if ( na <= 0.0 ) {
    *n = 0; 
    return 0.0;
  }
  cdev /= na;

  /* Total dispersion */
  _fff_imatch_marginalize( hI, H, clampI, clampJ, 0 ); 
  _fff_L1_moments ( hI, clampI, 1, &med, &dev, &na );

  
  /* Test on total dispersion */
  *n = na; 
  if ( dev==0.0 )
    return 0.0;
  else
    return( 1 - FFF_SQR(cdev)/FFF_SQR(dev) ); /* Squaring for comparison with CR(L2) */  
}



double fff_imatch_joint_ent( const double* H, int clampI, int clampJ )
{
  double n; 
  double entIJ = _fff_imatch_entropy( H, clampI*clampJ, &n ); 
  return entIJ; 
}


double fff_imatch_cond_ent( const double* H, double* hJ, int clampI, int clampJ )
{
  double n; 
  double entIJ, entJ;  
  _fff_imatch_marginalize( hJ, H, clampI, clampJ, 1 ); 
  entIJ = _fff_imatch_entropy( H, clampI*clampJ, &n ); 
  entJ = _fff_imatch_entropy( hJ, clampJ, &n ); 
  return( entIJ - entJ ); /* Entropy of I given J */ 
}


double fff_imatch_mi( const double* H, double* hI, int clampI, double* hJ, int clampJ )
{
  double n; 
  return( _fff_imatch_mi( H, hI, clampI, hJ, clampJ, &n) ); 
}

double fff_imatch_max_woe_mi( const double* H, double* hI, int clampI, double* hJ, int clampJ )
{
  double n, x; 
  x = _fff_imatch_mi( H, hI, clampI, hJ, clampJ, &n); 
  return( n*x ); 
}

static double _fff_imatch_mi( const double* H, double* hI, int clampI, double* hJ, int clampJ, double* n )
{
  double entIJ, entI, entJ; 
  _fff_imatch_marginalize( hI, H, clampI, clampJ, 0 ); 
  _fff_imatch_marginalize( hJ, H, clampI, clampJ, 1 ); 
  entI = _fff_imatch_entropy( hI, clampI, n ); 
  entJ = _fff_imatch_entropy( hJ, clampJ, n ); 
  entIJ = _fff_imatch_entropy( H, clampI*clampJ, n ); 
  return( entI + entJ - entIJ ); 
}

/*
  Normalized mutual information as advocated by Studholme, 98. 

  NMI = 2*(1 - H(I,J)/[H(I)+H(J)])
  
*/

double fff_imatch_nmi( const double* H, double* hI, int clampI, double* hJ, int clampJ )
{
  double n; 
  double entIJ, entI, entJ, aux; 
  _fff_imatch_marginalize( hI, H, clampI, clampJ, 0 ); 
  _fff_imatch_marginalize( hJ, H, clampI, clampJ, 1 ); 
  entI = _fff_imatch_entropy( hI, clampI, &n ); 
  entJ = _fff_imatch_entropy( hJ, clampJ, &n ); 
  entIJ = _fff_imatch_entropy( H, clampI*clampJ, &n );
  aux = entI + entJ; 
  if ( aux > 0.0 ) 
    return( 2*(1-entIJ/aux) );
  else 
    return 0.0; 
}


/*

Weight of evidence. 

*/ 

double fff_imatch_woe( const double* H, const double* F, 
		       double* fI, int clampI, double* fJ, int clampJ )
{
  double n; 
  return(  _fff_imatch_woe(H, F, fI, clampI, fJ, clampJ, &n) );   
}

static double _fff_imatch_woe( const double* H, const double* F, 
			       double* fI, int clampI, double* fJ, int clampJ, 
			       double* n )
{
  const double *bufH = H, *bufF = F; 
  double *buf_fI, *buf_fJ; 
  int i, j; 
  double hij, fij, fi, fj, aux, sumF, na = 0.0, woe = 0.0; 

  sumF = _fff_imatch_marginalize( fI, F, clampI, clampJ, 0 ); 
  sumF = _fff_imatch_marginalize( fJ, F, clampI, clampJ, 1 ); 

  for (i=0, buf_fI=fI; i<clampI; i++, buf_fI++) {
    fi = *buf_fI / sumF; /* HACK to implicitely normalize F */ 
    for (j=0, buf_fJ=fJ; j<clampJ; j++, buf_fJ++, bufH++, bufF++) {
      hij = *bufH;
      na += hij;  
      fj = *buf_fJ;
      fij = *bufF;

      /* If fi=0 or fj=0, then fij = 0. We conventionally set
	 fij/(fi*fj) to zero but this might be problematic */ 
      aux = fi * fj; 
      if ( aux > 0 ) 
	aux = fij / aux; 
      else 
	aux = 0.0;  
      woe += hij * NICELOG(aux); 
    }
  }

  *n = na;
  return woe; 
}



/* 
   First loop: compute the histogram sum 
   Second loop: compute actual entropy 
*/ 
static double _fff_imatch_entropy( const double* h, size_t size, double* n )
{
  double E=0.0, sumh=0.0, aux; 
  size_t i; 
  double *buf;

  buf = (double*)h; 
  for ( i=0; i<size; i++, buf++ ) 
    sumh += *buf; 

  if ( sumh <= 0 ) {
    *n = 0; 
    return 0.0; 
  }

  buf = (double*)h; 
  for ( i=0; i<size; i++, buf++ ) {
    aux = *buf / sumh; 
    if ( aux > 0 )
      E -= aux * log( aux ); 
  }

  *n = sumh; 
  return E; 
}






/*
  
Subsampling factors are determined by the speedup parameter of the
fff_match structure. We adopt the convention that a speedup factor of
n^3 corresponds to the situation where a 256^3 dimensional image is
subsampled with a factor n in each direction. Thus, an upper bound for
the number of voxels in the subsampled image is 256^3/speedup.

The following routine sets the subsampling factor so as to
reach this bound approximately.

*/
size_t fff_imatch_subsampling_factors( 
	    unsigned int* fX, 
	    unsigned int* fY, 
	    unsigned int* fZ,
	    unsigned int dimX, 
	    unsigned int dimY, 
	    unsigned int dimZ, 
	    double speedup )
{
  double ddx, ddy, ddz, nlim; 
  size_t nvox, dimXYZ = dimX*dimY*dimZ; 
  unsigned int dir; 

  *fX = *fY = *fZ = 1;
  nlim = SPEEDUP_REF / speedup;
  nvox = dimXYZ / ( (*fX)*(*fY)*(*fZ) );
  
  while ( nvox > nlim ) {
    
    /* Subsample the direction with the highest number of samples */
    ddx = (double)dimX / (double)*fX;
    ddy = (double)dimY / (double)*fY;
    ddz = (double)dimZ / (double)*fZ;
    
    if ( (ddx >= ddy) && (ddx >= ddz) )
      dir = 0;
    else {
      if ( (ddy > ddx) && (ddy >= ddz) )
	dir = 1;
      else
	dir = 2;
    }
    
    if ( dir == 0 ) (*fX) ++;
    if ( dir == 1 ) (*fY) ++;
    if ( dir == 2 ) (*fZ) ++;

    nvox = dimXYZ / ( (*fX)*(*fY)*(*fZ) );
  }
  
  return nvox; 
}



static void _fff_L1_moments ( const double * h, int clamp, int stride, 
			      double* median, double* dev, double* sumh )

{
  int i, med;
  double sum, cpdf, lim, auxdev;
  const double *buf;

  /* Initialisation au cas ou */
  *median=0; *dev=0; *sumh=0;
  cpdf = sum = 0;
  
  buf = h;
  for (i=0; i<clamp; i++, buf+=stride) {
    sum += *buf;
  }
  
  *sumh = sum;
  
  if (sum == 0)
    return;
  
  /* On cherche la valeur i telle que h(i-1) < sum/2 
     et h(i) >= sum/2 */
  lim = 0.5*sum;
  
  i = 0;
  buf = h;
  cpdf = *buf;
  auxdev = 0;
  
  while ( cpdf < lim ) {
    i ++;
    buf += stride;
    cpdf += *buf;
    auxdev += - i*(*buf);
  }
  
  /* 
     On a alors i-1 < med < i. On prend i comme valeur mediane, 
     meme s'il serait possible de choisir une valeur intermediaire
     entre i-1 et i, en approchant la fonction de repartition 
     par une droite. 
     
     La dispersion L1 est donnee par :
     
     sum*E(|X-med|) = - sum_{i<=med} i h(i)              [1]
     
                      + sum_{i>med} i h(i)               [2]
     
                      + med * [2*cpdf(med) - sum]        [3]


     Le terme [1] est actuellement egal a la variable auxdev.
     Le terme cpdf(med) est actuellement egal a la var. cpdf.
  */
  
  *median = (double)i;
  auxdev += (2*cpdf - sum)*(*median);
  med = i+1;
  
  
  /* Pour achever le calcul de la deviation L1, il suffit de calculer 
     la moyenne tronquee (entre i et la dimension) (cf. [2]) */
  
  if ( med < clamp ) {
    buf = h + med*stride;
    for (i=med; i<clamp; i ++, buf += stride) 
      auxdev += i*(*buf);
  }
  
  *dev = auxdev/sum; 

  return;           
}


/* =========================================================================
   RESAMPLING ROUTINES
   ========================================================================= */

/* Tvox is the voxel transformation from source to target 
   Resample a 2d-3d image undergoing an affine transformation. */
void fff_imatch_resample( fff_array* im_resampled, 
			  const fff_array* im, 
			  const double* Tvox )
{
  double i1;
  fff_array* im_spline_coeff;
  fff_array_iterator imIter = fff_array_iterator_init( im_resampled ); 
  size_t pos, x, y, z;
  size_t ddimX=im->dimX-1, ddimY=im->dimY-1, ddimZ=im->dimZ-1; 
  double Tx, Ty, Tz;
  const double* bufTvox; 
  fff_vector* work; 
  size_t work_size; 

  /* Compute the spline coefficient image */
  im_spline_coeff = fff_array_new3d( FFF_DOUBLE, im->dimX, im->dimY, im->dimZ );
  work_size = FFF_MAX( im->dimX, im->dimY ); 
  work_size = FFF_MAX( work_size, im->dimZ ); 
  work = fff_vector_new( work_size );   
  fff_cubic_spline_transform_image( im_spline_coeff, im, work );
  fff_vector_delete( work ); 

  /* Resampling loop */
  pos = 0; 
  while( imIter.uncomplete ) {
    x = imIter.x;
    y = imIter.y; 
    z = imIter.z; 
    APPLY_AFFINE_TRANSFORMATION(Tx, Ty, Tz, Tvox, bufTvox, x, y, z);     

    if ( (Tx<0) || (Tx>ddimX) ||
	 (Ty<0) || (Ty>ddimY) ||
	 (Tz<0) || (Tz>ddimZ) )
      i1 = 0.0; 
    else
      i1 = fff_cubic_spline_sample_image( Tx, Ty, Tz, 0, im_spline_coeff ); 

    fff_array_set3d(im_resampled, x, y, z, i1); 
    pos += fff_array_iterator_update( &imIter ); 
  }

  /* Free coefficient image */
  fff_array_delete( im_spline_coeff ); 

  return;
	 
}




/* =========================================================================
   MEMORY ALLOCATION
   ========================================================================= */

fff_imatch* fff_imatch_new ( const fff_array* imI,
			     const fff_array* imJ,
			     double thI,
			     double thJ,
			     int clampI, 
			     int clampJ )
{
  fff_imatch* imatch;
  
  /* Verify that input images are not 4D */ 
  if ( (imI->ndims == FFF_ARRAY_4D) ||
       (imJ->ndims == FFF_ARRAY_4D) ) {
    fff_warning("fff_imatch_new", "Input images cannot be 4D.\n" ); 
    return NULL; 
  }


  /* Start with allocating the object */
  imatch = (fff_imatch*)calloc(1, sizeof(fff_imatch));
  if (imatch == NULL) 
    return NULL;

  /** Create SSHORT images, clamp and possibly padd original images **/
  
  /* Source image */ 
  imatch->imI = fff_array_new3d( FFF_SSHORT, imI->dimX, imI->dimY, imI->dimZ );
  fff_array_clamp( imatch->imI, imI, thI, &clampI );
  

  /* Target image: enlarge to padd borders with negative values */ 
  imatch->imJ_padded = fff_array_new3d( FFF_SSHORT, imJ->dimX+2, imJ->dimY+2, imJ->dimZ+2 );
  fff_array_set_all( imatch->imJ_padded, -1 ); 
  imatch->imJ = (fff_array*)calloc(1, sizeof(fff_array) ); 
  *(imatch->imJ) = fff_array_get_block3d( imatch->imJ_padded, 
					    1, imJ->dimX, 1,  
					    1, imJ->dimY, 1,
					    1, imJ->dimZ, 1 );
  fff_array_clamp( imatch->imJ, imJ, thJ, &clampJ );
  

  /* Create the joint histogram structure. Important notice: in all
     computations, H will be assumed C-contiguous. 

     i (source intensities) are row indices
     j (target intensities) are column indices
  */ 
  imatch->clampI = clampI; 
  imatch->clampJ = clampJ; 
  imatch->H = calloc( clampI*clampJ, sizeof(double) ); 
  imatch->hI = calloc( clampI, sizeof(double) ); 
  imatch->hJ = calloc( clampJ, sizeof(double) ); 

  imatch->owner = 1;

  return imatch;
}

void fff_imatch_delete( fff_imatch* imatch )
{
  free( imatch->imJ ); /* This image is just a view on imJ_padded */ 
  if ( imatch->owner ) {
    fff_array_delete( imatch->imI ); 
    fff_array_delete( imatch->imJ_padded ); 
    free( imatch->H ); 
    free( imatch->hI ); 
    free( imatch->hJ ); 
  }
  free( imatch );
  return;
}
